# -*- coding: utf-8 -*-
# Copyright 2022 Aidentified LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import requests

import aidentified_matching_api.constants as constants
import aidentified_matching_api.get_id as get_id
import aidentified_matching_api.token_service as token


def list_datasets(args):
    resp_obj = token.token_service.paginated_api_call(
        args, requests.get, "/v1/dataset/"
    )
    constants.pretty(resp_obj)


def create_dataset(args):
    dataset_payload = {"name": args.name}
    resp_obj = token.token_service.api_call(
        args, requests.post, "/v1/dataset/", json=dataset_payload
    )

    constants.pretty(resp_obj)


def delete_dataset(args):
    args.dataset_name = args.name
    dataset_id = get_id.get_dataset_id_from_dataset_name(args)

    token.token_service.api_call(
        args,
        requests.delete,
        f"/v1/dataset/{dataset_id}/",
    )



# -*- coding: utf-8 -*-
# Copyright 2022 Aidentified LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import logging
import os
import pickle
import urllib.parse
import zlib

import appdirs
import requests

import aidentified_matching_api.constants as constants


logger = logging.getLogger("api")


class TokenService:
    __slots__ = ["expires_at", "token", "cache_file"]

    def __init__(self):
        self.expires_at = 0
        self.token = ""
        dirs = appdirs.AppDirs(
            appname="aidentified_match", appauthor="Aidentified", version="1.0"
        )
        os.makedirs(dirs.user_cache_dir, exist_ok=True)
        endpoint_hash = hex(zlib.crc32(constants.AIDENTIFIED_URL.encode("utf-8")))[2:]
        self.cache_file = os.path.join(
            dirs.user_cache_dir, f"token_cache_{endpoint_hash}"
        )

    def _read_token_cache(self):
        try:
            with open(self.cache_file, "rb") as fd:
                token_cache = pickle.load(fd)
                self.token = token_cache.get("token", "")
                self.expires_at = token_cache.get("expires_at", 0)
        except FileNotFoundError:
            pass

    def _write_token_cache(self):
        cache_value = {"token": self.token, "expires_at": self.expires_at}
        with open(self.cache_file, "wb") as fd:
            pickle.dump(cache_value, fd, protocol=pickle.HIGHEST_PROTOCOL)

    def get_token(self, args) -> str:
        self._read_token_cache()

        if datetime.datetime.utcnow().timestamp() < self.expires_at:
            return self.token

        # N.B. these are read from envvars AID_EMAIL and
        # AID_PASSWORD by default
        login_payload = {
            "email": args.email,
            "password": args.password,
        }

        logger.info("get_token /login")
        try:
            resp = requests.post(
                f"{constants.AIDENTIFIED_URL}/login", json=login_payload
            )
        except requests.exceptions.RequestException as e:
            raise Exception(f"Unable to connect to API: {e}") from None

        try:
            resp_payload = resp.json()
        except ValueError:
            raise Exception(f"Unable to parse API response: {resp.content}") from None

        try:
            resp.raise_for_status()
        except requests.exceptions.RequestException:
            raise Exception(
                f"Bad response from API: {resp.status_code} {resp_payload}"
            ) from None

        expires_at_dt = (
            datetime.timedelta(seconds=resp_payload["expires_in"])
            + datetime.datetime.utcnow()
        )

        self.expires_at = expires_at_dt.timestamp()
        self.token = resp_payload["bearer_token"]

        self._write_token_cache()

        return self.token

    def get_auth_headers(self, args) -> dict:
        return {"Authorization": f"Bearer {self.get_token(args)}"}

    def api_call(self, args, fn, url, **kwargs):
        auth_headers = self.get_auth_headers(args)

        if "headers" in kwargs:
            kwargs["headers"].update(auth_headers)
        else:
            kwargs["headers"] = auth_headers

        logger.info(f"{fn.__name__} {url}")
        try:
            resp: requests.Response = fn(f"{constants.AIDENTIFIED_URL}{url}", **kwargs)
        except requests.RequestException as e:
            raise Exception(f"Unable to make API call: {e}") from None

        if not resp.content:
            resp_obj = {}
        else:
            try:
                resp_obj = resp.json()
            except ValueError:
                raise Exception(
                    f"Unable to make API call: invalid response {resp.content}"
                ) from None

        try:
            resp.raise_for_status()
        except requests.RequestException:
            raise Exception(
                f"Unable to make API call: {resp.status_code} {resp_obj}"
            ) from None

        return resp_obj

    def paginated_api_call(self, args, fn, url, **kwargs):
        resp = []
        fetch_url = url
        parsed_orig_url = urllib.parse.urlparse(url)

        while True:
            paged = self.api_call(args, fn, fetch_url, **kwargs)
            resp.extend(paged["results"])
            if paged["next"] is None:
                break

            parsed_page_url = urllib.parse.urlparse(paged["next"])
            parsed_orig_url = parsed_orig_url._replace(query=parsed_page_url.query)
            fetch_url = parsed_orig_url.geturl()

        return resp


def get_token(args):
    if args.clear_cache:
        try:
            os.remove(token_service.cache_file)
        except FileNotFoundError:
            pass

    print(token_service.get_token(args))


token_service = TokenService()



# -*- coding: utf-8 -*-
# Copyright 2022 Aidentified LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import requests

import aidentified_matching_api.token_service as token


def get_dataset_id_from_dataset_name(args):
    dataset_params = {"name": args.dataset_name}
    resp_obj = token.token_service.api_call(
        args, requests.get, "/v1/dataset/", params=dataset_params
    )

    if resp_obj["count"] == 0:
        raise Exception(f"No dataset with name '{args.dataset_name}' found")

    return resp_obj["results"][0]["dataset_id"]


def get_dataset_file_id_from_dataset_file_name(args):
    dataset_params = {
        "dataset_name": args.dataset_name,
        "name": args.dataset_file_name,
    }
    resp_obj = token.token_service.api_call(
        args, requests.get, "/v1/dataset-file/", params=dataset_params
    )

    if resp_obj["count"] == 0:
        raise Exception(f"No dataset file with name '{args.dataset_file_name}' found")

    return resp_obj["results"][0]["dataset_file_id"]



# -*- coding: utf-8 -*-
# Copyright 2022 Aidentified LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import codecs
import csv
import io

import aidentified_matching_api.constants as constants

# first_name - string - required
# last_name - string - required
# id - string - optional; unique
# street_address_1 - string - optional; home street
# street_address_2 - string - optional; home street 2
# city - string - optional; home city
# state - string - optional; home state
# postal_code - string - optional; 5 digit postal code
# company - string - optional; name of the company of employment
# title - string - optional; job title
# school[_{n}] - string - optional; n in 1 to 10, name of the school of employment
# email[_{n}] - string - optional; n in 1 to 10
# phone[_{n}] - string - optional; n in 1 to 10, 10 digit phone formatted 5558675309
# domain[_{n}] - string - optional; n in 1 to 10, email domain such as "aidentified.com"
# linkedin[_{n}] - string - optional; n in 1 to 10


HEADERS = {
    "first_name",
    "last_name",
    "id",
    "street_address_1",
    "street_address_2",
    "city",
    "state",
    "postal_code",
    "company",
    "title",
    "school",
    "email",
    "phone",
    "domain",
    "linkedin",
    *[f"school_{idx}" for idx in range(1, 11)],
    *[f"email_{idx}" for idx in range(1, 11)],
    *[f"phone_{idx}" for idx in range(1, 11)],
    *[f"domain_{idx}" for idx in range(1, 11)],
    *[f"linkedin_{idx}" for idx in range(1, 11)],
}


class CsvArgs:
    __slots__ = [
        "raw_fd",
        "codec_info",
        "delimiter",
        "doublequotes",
        "escapechar",
        "quotechar",
        "quoting",
        "skipinitialspace",
    ]

    def __init__(
        self,
        raw_fd: io.BytesIO,
        codec_info: codecs.CodecInfo,
        delimiter: str,
        doublequotes: bool,
        escapechar: str,
        quotechar: str,
        quoting: int,
        skipinitialspace: bool,
    ):
        self.raw_fd = raw_fd
        self.codec_info = codec_info
        self.delimiter = delimiter.replace("\\t", "\t")
        self.doublequotes = doublequotes
        self.escapechar = escapechar
        self.quotechar = quotechar
        self.quoting = quoting
        self.skipinitialspace = skipinitialspace


def _csv_read(csv_reader, record_idx):
    try:
        return next(csv_reader)
    except csv.Error as e:
        raise Exception(f"Bad CSV format in row {record_idx}: {e}") from None
    except UnicodeError as e:
        raise Exception(f"Bad character encoding at byte {e.start}") from None


def validate(args) -> CsvArgs:
    # Validate choice of csv encoding, even if they don't do
    # the rest of the validation.
    try:
        codec_info = codecs.lookup(args.csv_encoding)
    except LookupError:
        raise Exception(f"Unknown csv-encoding '{args.csv_encoding}'") from None

    csv_args = CsvArgs(
        args.dataset_file_path,
        codec_info,
        args.csv_delimiter,
        args.csv_no_doublequotes,
        args.csv_escapechar,
        args.csv_quotechar,
        constants.QUOTE_METHODS[args.csv_quoting],
        args.csv_skip_initial_space,
    )

    if not args.validate:
        return csv_args

    validate_fd(csv_args)
    args.dataset_file_path.seek(0)

    return csv_args


def validate_fd(csv_args: CsvArgs):
    potential_bom = csv_args.raw_fd.read(3)
    # Python drops the BOM from UTF16/UTF32 but not UTF8
    if potential_bom == codecs.BOM_UTF8:
        skip_len = 3
    else:
        skip_len = 0
    csv_args.raw_fd.seek(0)
    csv_args.raw_fd.read(skip_len)

    text_fd = csv_args.codec_info.streamreader(csv_args.raw_fd)

    csv_reader = csv.reader(
        text_fd,
        delimiter=csv_args.delimiter,
        doublequote=csv_args.doublequotes,
        escapechar=csv_args.escapechar,
        quotechar=csv_args.quotechar,
        quoting=csv_args.quoting,
        skipinitialspace=csv_args.skipinitialspace,
        strict=True,
    )

    record_idx = 1

    try:
        headers = _csv_read(csv_reader, record_idx)
    except StopIteration:
        raise Exception("No headers in file") from None

    for header in headers:
        if header.lower() not in HEADERS:
            raise Exception(f"Invalid header '{header}'")

    sentinel = object()
    if (
        next(
            (
                header
                for header in headers
                if header not in ("id", "first_name", "last_name")
            ),
            sentinel,
        )
        is sentinel
    ):
        raise Exception("Needs at least one of the extra attribute headers")

    record_len = len(headers)

    try:
        id_idx = headers.index("id")
    except ValueError:
        id_idx = None
    id_uniqueness = set()

    required_header_idxes = []
    for required_header in ("first_name", "last_name"):
        if required_header not in headers:
            raise Exception(f"Required header {required_header} not in headers")

        required_header_idxes.append((required_header, headers.index(required_header)))

    record_idx += 1

    while True:
        try:
            record = _csv_read(csv_reader, record_idx)
        except StopIteration:
            break

        if record_idx > 500_001:
            raise Exception("CSV has more than 500,000 data rows")

        if len(record) != record_len:
            raise Exception(f"Row {record_idx} does not match header length")

        if id_idx is not None:
            cust_id = record[id_idx]
            if cust_id in id_uniqueness:
                raise Exception(f"Row {record_idx} has duplicate id '{cust_id}'")
            id_uniqueness.add(cust_id)

        for required_header, required_header_idx in required_header_idxes:
            if not record[required_header_idx]:
                raise Exception(
                    f"Row {record_idx} has invalid value for {required_header}"
                )

        record_idx += 1



# -*- coding: utf-8 -*-
# Copyright 2022 Aidentified LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import base64
import codecs
import contextlib
import csv
import functools
import hashlib
import io
import logging
import threading

import requests

import aidentified_matching_api.constants as constants
import aidentified_matching_api.get_id as get_id
import aidentified_matching_api.token_service as token
import aidentified_matching_api.validation as validation

logger = logging.getLogger("matching_api_cli")


UPLOAD_CANCELLED = threading.Event()


def list_dataset_files(args):
    dataset_file_params = {
        "dataset_name": args.dataset_name,
    }
    resp_obj = token.token_service.paginated_api_call(
        args, requests.get, "/v1/dataset-file/", params=dataset_file_params
    )
    constants.pretty(resp_obj)


def abort_dataset_file(args):
    dataset_file_id = get_id.get_dataset_file_id_from_dataset_file_name(args)
    resp_obj = token.token_service.api_call(
        args, requests.post, f"/v1/dataset-file/{dataset_file_id}/abort-upload/"
    )
    constants.pretty(resp_obj)


def create_dataset_file(args):
    # create files under name.
    dataset_id = get_id.get_dataset_id_from_dataset_name(args)

    dataset_file_payload = {"dataset_id": dataset_id, "name": args.dataset_file_name}
    resp_obj = token.token_service.api_call(
        args, requests.post, "/v1/dataset-file/", json=dataset_file_payload
    )

    constants.pretty(resp_obj)


async def file_uploader(args, dataset_file_id: str, part_queue: asyncio.Queue):
    loop = asyncio.get_event_loop()

    while True:
        part_idx, part_data = await part_queue.get()
        aws_part_number = part_idx + 1

        logger.info(f"Starting upload part {aws_part_number} hash")
        md5 = await loop.run_in_executor(
            None, lambda data: base64.b64encode(hashlib.md5(data).digest()), part_data
        )

        upload_part_payload = {
            "dataset_file_id": dataset_file_id,
            "part_number": aws_part_number,
            "md5": md5.decode("UTF-8"),
        }
        upload_part_callable = functools.partial(
            token.token_service.api_call,
            args,
            requests.post,
            "/v1/dataset-file-upload-part/",
            json=upload_part_payload,
        )
        resp = await loop.run_in_executor(None, upload_part_callable)
        upload_url = resp["upload_url"]
        dataset_file_upload_part_id = resp["dataset_file_upload_part_id"]

        logger.info(f"Starting upload part {aws_part_number} upload")
        put_part_callable = functools.partial(
            requests.put,
            upload_url,
            data=part_data,
            headers={"content-md5": md5},
        )
        try:
            upload_resp = await loop.run_in_executor(None, put_part_callable)
        except requests.exceptions.RequestException as e:
            raise Exception(f"Unable to upload file part: {e}") from None

        try:
            upload_resp.raise_for_status()
        except requests.exceptions.RequestException:
            # S3 returns XML. If it fails, let's just spew it.
            raise Exception(
                f"Unable to upload file part: {upload_resp.status_code} {upload_resp.text}"
            ) from None

        patch_etag_callable = functools.partial(
            token.token_service.api_call,
            args,
            requests.patch,
            f"/v1/dataset-file-upload-part/{dataset_file_upload_part_id}/",
            json={"etag": upload_resp.headers["ETag"]},
        )
        await loop.run_in_executor(None, patch_etag_callable)

        part_queue.task_done()
        logger.info(f"Finished upload part {aws_part_number}")


@contextlib.contextmanager
def upload_abort_ctxmgr(args, dataset_file_id: str):
    try:
        yield
    except:  # noqa: E722
        token.token_service.api_call(
            args, requests.post, f"/v1/dataset-file/{dataset_file_id}/abort-upload/"
        )
        raise


SENTINEL = object()


async def rewrite_csv(
    csv_args: validation.CsvArgs, part_size_bytes: int, part_queue: asyncio.Queue
):
    loop = asyncio.get_event_loop()

    part_idx = 0
    utf_8_info = codecs.lookup("UTF-8")

    out_buf = b""
    out_bytes_fd = io.BytesIO()
    out_text_fd = utf_8_info.streamwriter(out_bytes_fd)

    read_fd = csv_args.codec_info.streamreader(csv_args.raw_fd)

    reader = csv.reader(
        read_fd,
        delimiter=csv_args.delimiter,
        doublequote=csv_args.doublequotes,
        escapechar=csv_args.escapechar,
        quotechar=csv_args.quotechar,
        quoting=csv_args.quoting,
        skipinitialspace=csv_args.skipinitialspace,
        strict=True,
    )
    writer = csv.writer(out_text_fd, quoting=csv.QUOTE_MINIMAL)

    while True:
        row = await loop.run_in_executor(None, next, reader, SENTINEL)
        if row is SENTINEL:
            out_buf += out_bytes_fd.getvalue()
            break

        writer.writerow(row)

        if out_bytes_fd.tell() < part_size_bytes:
            continue

        out_buf += out_bytes_fd.getvalue()

        out_bytes_fd.seek(0)
        out_bytes_fd.truncate()

        logger.info(f"Putting upload part {part_idx+1}")
        await part_queue.put((part_idx, out_buf[:part_size_bytes]))
        out_buf = out_buf[part_size_bytes:]
        part_idx += 1

    if len(out_buf) > 0:
        logger.info(f"Putting final upload part {part_idx+1}")
        await part_queue.put((part_idx, out_buf))


async def manage_uploads(args, dataset_file_id: str, csv_args: validation.CsvArgs):
    part_queue = asyncio.Queue(maxsize=args.concurrent_uploads)
    part_size_bytes = args.upload_part_size * 1024 * 1024

    uploader_tasks = []

    for _ in range(args.concurrent_uploads):
        uploader_tasks.append(
            asyncio.create_task(file_uploader(args, dataset_file_id, part_queue))
        )

    async def part_queue_joiner():
        await rewrite_csv(csv_args, part_size_bytes, part_queue)
        # now that everything is queued, join() for work to finish
        await part_queue.join()

        # now that everything is done, cancel() the otherwise idle workers
        for uploader_task in uploader_tasks:
            uploader_task.cancel()

    coros = [part_queue_joiner(), *uploader_tasks]

    # await on all tasks in case any of them raise an exception, so you
    # can kill them all
    done, pending = await asyncio.wait(coros, return_when=asyncio.FIRST_EXCEPTION)

    # if pending, an exception hit us
    for pending_fut in pending:
        pending_fut.cancel()

    had_exception = [
        future.exception()
        for future in done
        if not future.cancelled() and future.exception() is not None
    ]
    if had_exception:
        exc_strings = ", ".join(
            f"Task {fut_idx}: {exc}" for fut_idx, exc in enumerate(had_exception)
        )
        raise Exception(f"Error(s) while uploading file: {exc_strings}")


def upload_dataset_file(args):
    if args.upload_part_size < 5:
        raise Exception("--upload-part-size must be greater than 5 Mb")

    logger.info("Starting validation")
    csv_args = validation.validate(args)
    logger.info("Validation complete")

    dataset_file_id = get_id.get_dataset_file_id_from_dataset_file_name(args)

    token.token_service.api_call(
        args, requests.post, f"/v1/dataset-file/{dataset_file_id}/initiate-upload/"
    )

    loop = asyncio.new_event_loop()
    with upload_abort_ctxmgr(args, dataset_file_id):
        loop.run_until_complete(manage_uploads(args, dataset_file_id, csv_args))

    complete_resp = token.token_service.api_call(
        args, requests.post, f"/v1/dataset-file/{dataset_file_id}/complete-upload/"
    )
    constants.pretty(complete_resp)


def download_dataset_file(args):
    dataset_file_id = get_id.get_dataset_file_id_from_dataset_file_name(args)

    resp_obj = token.token_service.api_call(
        args, requests.get, f"/v1/dataset-file/{dataset_file_id}/"
    )
    if resp_obj["download_url"] is None:
        raise Exception("Dataset file is not ready for download.")

    try:
        download_req = requests.get(resp_obj["download_url"])
    except requests.exceptions.RequestException as e:
        raise Exception(f"Unable to download file: {e}") from None

    for chunk in download_req.iter_content(chunk_size=1024 * 1024 * 1024):
        args.dataset_file_path.write(chunk)

    args.dataset_file_path.close()


def delete_dataset_file(args):
    dataset_file_id = get_id.get_dataset_file_id_from_dataset_file_name(args)

    token.token_service.api_call(
        args, requests.delete, f"/v1/dataset-file/{dataset_file_id}/"
    )
